/*
 * Copyright 2012 Steven Gribble
 *
 *  This file is part of the UW CSE333 project sequence (333proj12).
 *
 *  333proj12 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  333proj12 is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj12.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _HW2_JSONPARSER_H_
#define _HW2_JSONPARSER_H_

// JSON stands for "JavaScript Object Notation" -- it is a lightweight,
// easy-to-parse and easy-to-generate format for exchanging partially
// structured data.  Many Web services use JSON to exchange data with
// clients through their APIs.  For details on the JSON format, see:
//
//     http://www.json.org/
//  or
//     http://en.wikipedia.org/wiki/JSON

#include <string>
#include <iostream>

// We need to do this since we'll be linking a library compiled for
// C by gcc with our new C++ code that is compiled by g++.  'extern "C"'
// tells the compiler that anything declared within that block was
// compiled with gcc, and therefore we should use C calling conventions
// instead of C++ conventions when accessing it.
extern "C" {
#include "./libhw1/LinkedList.h"
#include "./libhw1/AVLTree.h"
}

#include "./Utils.h"

// A JSONValue is an abstract base class that is the superclass of several
// kinds JSON values.  For example, descendent classes might be a JSON string,
// a JSON number, a JSON array, or a JSON object (which is kind of like a
// hash table).
//
// A JSONValue can be a nested structure.  For example, an Object can contain
// an array, and that array can contain other values.
//
// Any strings stored in a JSONValue are unescaped.
class JSONValue {
 public:
  JSONValue() { }

  // Pure virtual; forces descendants to define their own destructors.
  virtual ~JSONValue() = 0;

  // A class method that parses a UTF-8 encoded, NULL terminated C
  // string containing JSON encoded, escaped data, and allocates and
  // returns a JSONValuePtr value with the parsed data.
  //
  // Arguments:
  //
  // - data: the UTF-8 encoded string with the (escaped) JSON data
  //
  // Returns:
  //
  // - NULL on error (e.g., out of memory or parse failure)
  //
  // - a JSONValue* containing the parsed data on success.
  //   Caller takes ownership of this and must call delete
  //   to free it.
  static JSONValue *JSONParse(const std::string &data);

  // This dump method gets invoked by the overloaded "<<"
  // operator, and therefore allows for polymorphic overloading.
  virtual std::ostream& Dump(std::ostream &o) const = 0;

 private:
  DISALLOW_COPY_AND_ASSIGN(JSONValue);
};

// Overload the "<<" operator to allow a JSONValue (and descendants)
// to be written into a stream.
std::ostream& operator<<(std::ostream &o, const JSONValue &b);

// A JSONObject represents a map from string to value, i.e., a JavaScript
// object.  An example of a JSONObject represented in string form is:
//
//   { name: "Steve", job: "Professor", classes: [ "CSE333", "CSE451" ] }
class JSONObject : public JSONValue {
 public:
  JSONObject();
  ~JSONObject();

  // Returns the JSONValue associated with key "key".
  //
  // Arguments:
  //
  // - key: a UTF-8 encoded string that is the key to look up
  //   within the object.  The key should not be escaped or quoted.
  //
  // Returns:
  //
  // - NULL if the key does not exist within the object.
  //
  // - otherwise, a JSONValue* to the value associated with 'key'
  //   The caller does *not* get memory ownership of the value;
  //   whoever currently has ownership retains it.
  JSONValue *Get(const std::string &key);

  // Adds the value "value_to_add" under the key "key".
  //
  // Arguments:
  //
  // - key: a UTF-8 encoded, NULL terminated string of the
  //   key to add.  The key should not be escaped.
  //
  // - value_to_add:  the value to add to under key "key"
  //
  // Returns:
  //
  // - false:  if there is a a memory allocation error.
  //
  // - true: if "value_to_add" is successfully added.  If so,
  //   then the caller yields memory ownership of "value_to_add."
  //   If "key" already exists, then the existing value is replaced
  //   with "value_to_add", and that existing value is deleted.
  bool Put(const std::string &key, JSONValue *value_to_add);

  // Dumps an escaped string representation of the object.
  std::ostream& Dump(std::ostream &o) const;

  // This structure defines the payload of an element of the AVLtree
  // that we use to represent a JSON "object" type.  Each element of
  // the tree contains a std::string for the key and a new'ed
  // JSONValuePtr for the value.
  typedef struct object_tree_st {
    std::string  key;    // the key associated with the value
    JSONValue   *value;  // the value
  } ObjectTreeStruct, *ObjectTreeStructPtr;

  // Make the tree itself accessible so clients can iterate through it.
  AVLTreePtr tree_;
 private:

  DISALLOW_COPY_AND_ASSIGN(JSONObject);
};

// A JSONArray represents an array of JSONValues.  An example of a
// JSONArray represented in string form is:
//
//  [ "Steve", "is", "a", { professor: true, programmer: true } ]
class JSONArray : public JSONValue {
 public:
  JSONArray();
  ~JSONArray();

  // Overload the [] operator to make this feel array-like.
  // Returns the JSONValue associated with the value at offset
  // "index" within the array "array", or throws an exception
  // if the request index is past the end of the array.
  JSONValue *operator[](const uint64_t index);

  // Appends the value "value_to_append".
  //
  // Arguments:
  //
  // - value_to_append: the value to append to the array.
  //
  // Returns:
  //
  // - false: if there is a memory allocation error.
  //
  // - true if "value_to_append" is appended.  If so, caller
  //   yields memory ownership of "value_to_append", since
  //   it is incorporated into "array".
  bool Append(JSONValue *value_to_append);

  // Returns the length of the array.
  uint64_t Length();

  // Dumps an escaped string representation of the array.
  std::ostream& Dump(std::ostream &o) const;

 private:
  LinkedListPtr list_;

  DISALLOW_COPY_AND_ASSIGN(JSONArray);
};

// A JSONString represents a string within a JSON structure.
class JSONString : public JSONValue {
 public:
  JSONString() { }

  // Note: the string passed in should note be escaped.
  explicit JSONString(std::string str) : str_(str) { }
  ~JSONString() { }

  // Dumps an escaped string representation of the string.
  std::ostream& Dump(std::ostream &o) const;

  // Don't bother with accessors/setters, just allow direct
  // access to my internal value.
  std::string str_;

 private:
  DISALLOW_COPY_AND_ASSIGN(JSONString);
};

// A JSONNumber represents a floating point number.
class JSONNumber : public JSONValue {
 public:
  JSONNumber() { }
  explicit JSONNumber(double num) : num_(num) { }
  ~JSONNumber() { }

  // Dumps an escaped string representation of the number.
  std::ostream& Dump(std::ostream &o) const;

  // Don't bother with accessors/setters, just allow direct
  // access to my internal value.
  double num_;

 private:
  DISALLOW_COPY_AND_ASSIGN(JSONNumber);
};

// A JSONBoolean represents a true or false value.
class JSONBoolean : public JSONValue {
 public:
  JSONBoolean() { }
  explicit JSONBoolean(bool val) : val_(val) { }
  ~JSONBoolean() { }

  // Dumps an escaped string representation of the boolean.
  std::ostream& Dump(std::ostream &o) const;

  // Don't bother with accessors/setters, just allow direct
  // access to my internal value.
  bool val_;

 private:
  DISALLOW_COPY_AND_ASSIGN(JSONBoolean);
};

// A JSONNull represents the absence of a value, i.e. "null".
class JSONNull : public JSONValue {
 public:
  JSONNull() { }
  ~JSONNull() { }

  // Dumps an escaped string representation of the null.
  std::ostream& Dump(std::ostream &o) const;

 private:
  DISALLOW_COPY_AND_ASSIGN(JSONNull);
};

#endif  // _HW2_JSONPARSER_H_
